# encoding: utf-8
"""
reactor.py

Created by Thomas Mangin on 2012-06-10.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import re
import sys
import time
import signal
import select
import socket

from collections import deque

from exabgp.protocol.ip import IP

from exabgp.reactor.daemon import Daemon
from exabgp.reactor.listener import Listener
from exabgp.reactor.listener import NetworkError
from exabgp.reactor.api.processes import Processes
from exabgp.reactor.api.processes import ProcessError
from exabgp.reactor.peer import Peer
from exabgp.reactor.peer import ACTION
from exabgp.reactor.network.error import error

from exabgp.reactor.api.decoder import Decoder
from exabgp.configuration.ancient import Configuration
from exabgp.configuration.environment import environment

from exabgp.version import version
from exabgp.logger import Logger


class Reactor (object):
	# [hex(ord(c)) for c in os.popen('clear').read()]
	clear = ''.join([chr(int(c,16)) for c in ['0x1b', '0x5b', '0x48', '0x1b', '0x5b', '0x32', '0x4a']])

	def __init__ (self, configurations):
		self.ip = environment.settings().tcp.bind
		self.port = environment.settings().tcp.port
		self.respawn = environment.settings().api.respawn

		self.max_loop_time = environment.settings().reactor.speed
		self.early_drop = environment.settings().daemon.drop

		self.logger = Logger()
		self.daemon = Daemon(self)
		self.processes = None
		self.listener = None
		self.configuration = Configuration(configurations)
		self.decoder = Decoder()

		self.peers = {}
		self.route_update = False

		self._stopping = environment.settings().tcp.once
		self._shutdown = False
		self._reload = False
		self._reload_processes = False
		self._restart = False
		self._saved_pid = False
		self._pending = deque()
		self._running = None

		signal.signal(signal.SIGTERM, self.sigterm)
		signal.signal(signal.SIGHUP, self.sighup)
		signal.signal(signal.SIGALRM, self.sigalrm)
		signal.signal(signal.SIGUSR1, self.sigusr1)
		signal.signal(signal.SIGUSR2, self.sigusr2)

	def sigterm (self, signum, frame):
		self.logger.reactor("SIG TERM received - shutdown")
		self._shutdown = True

	def sighup (self, signum, frame):
		self.logger.reactor("SIG HUP received - shutdown")
		self._shutdown = True

	def sigalrm (self, signum, frame):
		self.logger.reactor("SIG ALRM received - restart")
		self._restart = True

	def sigusr1 (self, signum, frame):
		self.logger.reactor("SIG USR1 received - reload configuration")
		self._reload = True

	def sigusr2 (self, signum, frame):
		self.logger.reactor("SIG USR2 received - reload configuration and processes")
		self._reload = True
		self._reload_processes = True

	def ready (self, sockets, ios, sleeptime=0):
		# never sleep a negative number of second (if the rounding is negative somewhere)
		# never sleep more than one second (should the clock time change during two time.time calls)
		sleeptime = min(max(0.0,sleeptime),1.0)
		if not ios:
			time.sleep(sleeptime)
			return []
		try:
			read,_,_ = select.select(sockets+ios,[],[],sleeptime)
			return read
		except select.error,exc:
			errno,message = exc.args  # pylint: disable=W0633
			if errno not in error.block:
				raise exc
			return []
		except socket.error,exc:
			if exc.errno in error.fatal:
				raise exc
			return []
		except ValueError,exc:
			# pulled on a closed FD
			return []
		except KeyboardInterrupt,exc:
			raise exc

	def run (self):
		self.daemon.daemonise()

		# Make sure we create processes once we have closed file descriptor
		# unfortunately, this must be done before reading the configuration file
		# so we can nto do it with dropped privileges
		self.processes = Processes(self)

		# we have to read the configuration possibly with root privileges
		# as we need the MD5 information when we bind, and root is needed
		# to bind to a port < 1024

		# this is undesirable as :
		# - handling user generated data as root should be avoided
		# - we may not be able to reload the configuration once the privileges are dropped

		# but I can not see any way to avoid it

		if not self.load():
			return False

		try:
			self.listener = Listener()

			if self.ip:
				self.listener.listen(IP.create(self.ip),IP.create('0.0.0.0'),self.port,None)
				self.logger.reactor("Listening for BGP session(s) on %s:%d" % (self.ip,self.port))

			for neighbor in self.configuration.neighbor.values():
				if neighbor.listen:
					self.listener.listen(neighbor.local_address,neighbor.peer_address,neighbor.listen,neighbor.md5)
					self.logger.reactor("Listening for BGP session(s) on %s:%d%s" % (neighbor.local_address,neighbor.listen,' with MD5' if neighbor.md5 else ''))
		except NetworkError,exc:
			self.listener = None
			if os.geteuid() != 0 and self.port <= 1024:
				self.logger.reactor("Can not bind to %s:%d, you may need to run ExaBGP as root" % (self.ip,self.port),'critical')
			else:
				self.logger.reactor("Can not bind to %s:%d (%s)" % (self.ip,self.port,str(exc)),'critical')
			self.logger.reactor("unset exabgp.tcp.bind if you do not want listen for incoming connections",'critical')
			self.logger.reactor("and check that no other daemon is already binding to port %d" % self.port,'critical')
			sys.exit(1)

		if not self.early_drop:
			self.processes.start()

		if not self.daemon.drop_privileges():
			self.logger.reactor("Could not drop privileges to '%s' refusing to run as root" % self.daemon.user,'critical')
			self.logger.reactor("Set the environmemnt value exabgp.daemon.user to change the unprivileged user",'critical')
			return

		if self.early_drop:
			self.processes.start()

		# This is required to make sure we can write in the log location as we now have dropped root privileges
		if not self.logger.restart():
			self.logger.reactor("Could not setup the logger, aborting",'critical')
			return

		if not self.daemon.savepid():
			return

		# did we complete the run of updates caused by the last SIGUSR1/SIGUSR2 ?
		reload_completed = True

		wait = environment.settings().tcp.delay
		if wait:
			sleeptime = (wait * 60) - int(time.time()) % (wait * 60)
			self.logger.reactor("waiting for %d seconds before connecting" % sleeptime)
			time.sleep(float(sleeptime))

		workers = {}
		peers = set()
		scheduled = False

		while True:
			try:
				finished = False
				start = time.time()
				end = start + self.max_loop_time

				if self._shutdown:
					self._shutdown = False
					self.shutdown()
					break

				if self._reload and reload_completed:
					self._reload = False
					self.load()
					self.processes.start(self._reload_processes)
					self._reload_processes = False
				elif self._restart:
					self._restart = False
					self.restart()

				# We got some API routes to announce
				if self.route_update:
					self.route_update = False
					self.route_send()

				for peer in self.peers.keys():
					peers.add(peer)

				while start < time.time() < end and not finished:
					if self.peers:
						for key in list(peers):
							peer = self.peers[key]
							action = peer.run()

							# .run() returns an ACTION enum:
							# * immediate if it wants to be called again
							# * later if it should be called again but has no work atm
							# * close if it is finished and is closing down, or restarting
							if action == ACTION.CLOSE:
								self.unschedule(peer)
								peers.discard(key)
							# we are loosing this peer, not point to schedule more process work
							elif action == ACTION.LATER:
								for io in peer.sockets():
									workers[io] = key
								# no need to come back to it before a a full cycle
								peers.discard(key)

					if not peers:
						reload_completed = True

					if self.listener:
						for connection in self.listener.connected():
							# found
							# * False, not peer found for this TCP connection
							# * True, peer found
							# * None, conflict found for this TCP connections
							found = False
							for key in self.peers:
								peer = self.peers[key]
								neighbor = peer.neighbor
								# XXX: FIXME: Inet can only be compared to Inet
								if connection.local == str(neighbor.peer_address) and connection.peer == str(neighbor.local_address):
									if peer.incoming(connection):
										found = True
										break
									found = None
									break

							if found:
								self.logger.reactor('accepted connection from  %s - %s' % (connection.local,connection.peer))
							elif found is False:
								self.logger.reactor('no session configured for  %s - %s' % (connection.local,connection.peer))
								connection.notification(6,3,'no session configured for the peer')
								connection.close()
							elif found is None:
								self.logger.reactor('connection refused (already connected to the peer) %s - %s' % (connection.local,connection.peer))
								connection.notification(6,5,'could not accept the connection')
								connection.close()

					scheduled = self.schedule()
					finished = not peers and not scheduled

				# RFC state that we MUST not send more than one KEEPALIVE / sec
				# And doing less could cause the session to drop

				if finished:
					for io in self.ready(list(peers),self.processes.fds(),end-time.time()):
						if io in workers:
							peers.add(workers[io])
							del workers[io]

				if self._stopping and not self.peers.keys():
					break

			except KeyboardInterrupt:
				while True:
					try:
						self._shutdown = True
						self.logger.reactor("^C received")
						break
					except KeyboardInterrupt:
						pass
			except SystemExit:
				while True:
					try:
						self._shutdown = True
						self.logger.reactor("exiting")
						break
					except KeyboardInterrupt:
						pass
			# socket.error is a subclass of IOError (so catch it first)
			except socket.error:
				try:
					self._shutdown = True
					self.logger.reactor('socket error received','warning')
					break
				except KeyboardInterrupt:
					pass
			except IOError:
				while True:
					try:
						self._shutdown = True
						self.logger.reactor("I/O Error received, most likely ^C during IO",'warning')
						break
					except KeyboardInterrupt:
						pass
			except ProcessError:
				while True:
					try:
						self._shutdown = True
						self.logger.reactor("Problem when sending message(s) to helper program, stopping",'error')
						break
					except KeyboardInterrupt:
						pass
			except select.error:
				while True:
					try:
						self._shutdown = True
						self.logger.reactor("problem using select, stopping",'error')
						break
					except KeyboardInterrupt:
						pass
				# from exabgp.leak import objgraph
				# print objgraph.show_most_common_types(limit=20)
				# import random
				# obj = objgraph.by_type('Route')[random.randint(0,2000)]
				# objgraph.show_backrefs([obj], max_depth=10)

	def shutdown (self):
		"""terminate all the current BGP connections"""
		self.logger.reactor("Performing shutdown")
		if self.listener:
			self.listener.stop()
			self.listener = None
		for key in self.peers.keys():
			self.peers[key].stop()
		self.processes.terminate()
		self.daemon.removepid()
		self._stopping = True

	def load (self):
		"""reload the configuration and send to the peer the route which changed"""
		self.logger.reactor("Performing reload of exabgp %s" % version)

		reloaded = self.configuration.reload()

		if not reloaded:
			#
			# Careful the string below is used but the QA code to check for sucess of failure
			self.logger.configuration("Problem with the configuration file, no change done",'error')
			# Careful the string above is used but the QA code to check for sucess of failure
			#
			self.logger.configuration(self.configuration.error,'error')
			return False

		for key, peer in self.peers.items():
			if key not in self.configuration.neighbor:
				self.logger.reactor("Removing peer: %s" % peer.neighbor.name())
				peer.stop()

		for key, neighbor in self.configuration.neighbor.items():
			# new peer
			if key not in self.peers:
				self.logger.reactor("New peer setup: %s" % neighbor.name())
				peer = Peer(neighbor,self)
				self.peers[key] = peer
			# modified peer
			elif self.peers[key].neighbor != neighbor:
				self.logger.reactor("Peer definition change, establishing a new connection for %s" % str(key))
				self.peers[key].reestablish(neighbor)
			# same peer but perhaps not the routes
			else:
				# finding what route changed and sending the delta is not obvious
				self.logger.reactor("Peer definition identical, updating peer routes if required for %s" % str(key))
				self.peers[key].reconfigure(neighbor)
		self.logger.configuration("Loaded new configuration successfully",'info')

		return True

	def schedule (self):
		try:
			# read at least on message per process if there is some and parse it
			for service,command in self.processes.received():
				self.decoder.parse_command(self,service,command)

			# if we have nothing to do, return or save the work
			if not self._running:
				if not self._pending:
					return False
				self._running = self._pending.popleft()

			# run it
			try:
				self._running.next()  # run
				# should raise StopIteration in most case
				# and prevent us to have to run twice to run one command
				self._running.next()  # run
			except StopIteration:
				self._running = None
			return True

		except StopIteration:
			pass
		except KeyboardInterrupt:
			self._shutdown = True
			self.logger.reactor("^C received",'error')

	def route_send (self):
		"""the process ran and we need to figure what routes to changes"""
		self.logger.reactor("Performing dynamic route update")
		for key in self.configuration.neighbor.keys():
			self.peers[key].send_new()
		self.logger.reactor("Updated peers dynamic routes successfully")

	def restart (self):
		"""kill the BGP session and restart it"""
		self.logger.reactor("Performing restart of exabgp %s" % version)
		self.configuration.reload()

		for key in self.peers.keys():
			if key not in self.configuration.neighbor.keys():
				neighbor = self.configuration.neighbor[key]
				self.logger.reactor("Removing Peer %s" % neighbor.name())
				self.peers[key].stop()
			else:
				self.peers[key].reestablish()
		self.processes.terminate()
		self.processes.start()

	def unschedule (self, peer):
		key = peer.neighbor.name()
		if key in self.peers:
			del self.peers[key]

	def answer (self, service, string):
		self.processes.write(service,string)
		self.logger.reactor('Responding to %s : %s' % (service,string))

	def api_shutdown (self):
		self._shutdown = True
		self._pending = deque()
		self._running = None

	def api_reload (self):
		self._reload = True
		self._pending = deque()
		self._running = None

	def api_restart (self):
		self._restart = True
		self._pending = deque()
		self._running = None

	@staticmethod
	def match_neighbor (description, name):
		for string in description:
			if re.search(r'(^|[\s])%s($|[\s,])' % re.escape(string), name) is None:
				return False
		return True

	def match_neighbors (self, descriptions):
		"""return the sublist of peers matching the description passed, or None if no description is given"""
		if not descriptions:
			return self.peers.keys()

		returned = []
		for key in self.peers:
			for description in descriptions:
				if Reactor.match_neighbor(description,key):
					if key not in returned:
						returned.append(key)
		return returned

	def nexthops (self, peers):
		return dict((peer,self.peers[peer].neighbor.local_address) for peer in peers)

	def plan (self, callback):
		self._pending.append(callback)



# encoding: utf-8
"""
update/__init__.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import NoIP
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message import Message
from exabgp.bgp.message import IN
from exabgp.bgp.message import OUT
from exabgp.bgp.message.update.eor import EOR

from exabgp.bgp.message.update.attribute.attributes import Attributes
from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.attribute.mprnlri import MPRNLRI
from exabgp.bgp.message.update.attribute.mprnlri import EMPTY_MPRNLRI
from exabgp.bgp.message.update.attribute.mpurnlri import MPURNLRI
from exabgp.bgp.message.update.attribute.mpurnlri import EMPTY_MPURNLRI

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.nlri import NLRI

from exabgp.logger import Logger
from exabgp.logger import LazyFormat

# ======================================================================= Update

# +-----------------------------------------------------+
# |   Withdrawn Routes Length (2 octets)                |
# +-----------------------------------------------------+
# |   Withdrawn Routes (variable)                       |
# +-----------------------------------------------------+
# |   Total Path Attribute Length (2 octets)            |
# +-----------------------------------------------------+
# |   Path Attributes (variable)                        |
# +-----------------------------------------------------+
# |   Network Layer Reachability Information (variable) |
# +-----------------------------------------------------+

# Withdrawn Routes:

# +---------------------------+
# |   Length (1 octet)        |
# +---------------------------+
# |   Prefix (variable)       |
# +---------------------------+


class Update (Message):
	ID = Message.CODE.UPDATE
	TYPE = chr(Message.CODE.UPDATE)
	EOR = False

	def __init__ (self, nlris, attributes):
		self.nlris = nlris
		self.attributes = attributes

	# message not implemented we should use messages below.

	def __str__ (self):
		return '\n'.join(['%s%s' % (str(self.nlris[n]),str(self.attributes)) for n in range(len(self.nlris))])

	@staticmethod
	def prefix (data):
		return '%s%s' % (pack('!H',len(data)),data)

	@staticmethod
	def split (data):
		length = len(data)

		len_withdrawn = unpack('!H',data[0:2])[0]
		withdrawn = data[2:len_withdrawn+2]

		if len(withdrawn) != len_withdrawn:
			raise Notify(3,1,'invalid withdrawn routes length, not enough data available')

		start_attributes = len_withdrawn+4
		len_attributes = unpack('!H',data[len_withdrawn+2:start_attributes])[0]
		start_announced = len_withdrawn+len_attributes+4
		attributes = data[start_attributes:start_announced]
		announced = data[start_announced:]

		if len(attributes) != len_attributes:
			raise Notify(3,1,'invalid total path attribute length, not enough data available')

		if 2 + len_withdrawn + 2 + len_attributes + len(announced) != length:
			raise Notify(3,1,'error in BGP message length, not enough data for the size announced')

		return withdrawn,attributes,announced

	# The routes MUST have the same attributes ...
	# XXX: FIXME: calculate size progressively to not have to do it every time
	# XXX: FIXME: we could as well track when packed_del, packed_mp_del, etc
	# XXX: FIXME: are emptied and therefore when we can save calculations
	def messages (self, negotiated):
		# sort the nlris

		add_nlri = []
		del_nlri = []
		add_mp = {}
		del_mp = {}

		for nlri in self.nlris:
			if nlri.family() in negotiated.families:
				if nlri.afi == AFI.ipv4 and nlri.safi in [SAFI.unicast, SAFI.multicast]:
					if nlri.action == OUT.ANNOUNCE:
						add_nlri.append(nlri)
					else:
						del_nlri.append(nlri)
				else:
					if nlri.action == OUT.ANNOUNCE:
						add_mp.setdefault(nlri.family(),[]).append(nlri)
					else:
						del_mp.setdefault(nlri.family(),[]).append(nlri)

		if not add_nlri and not del_nlri and not add_mp and not del_mp:
			return

		if add_nlri or add_mp:
			attr = self.attributes.pack(negotiated,True)
		else:
			attr = ''

		# withdrawn IPv4

		packed_del = ''
		msg_size = negotiated.msg_size - 19 - 2 - 2  # 2 bytes for each of the two prefix() header
		addpath = negotiated.addpath.send(AFI.ipv4,SAFI.unicast)

		while del_nlri:
			nlri = del_nlri.pop()
			packed = nlri.pack(addpath)
			seen_size = len(packed_del + packed)
			if seen_size > msg_size:
				if not packed_del:
					raise Notify(6,0,'attributes size is so large we can not even pack one NLRI')
				yield self._message(Update.prefix(packed_del) + Update.prefix(''))
				packed_del = packed
			else:
				packed_del += packed

		# withdrawn MP

		packed_mp_del = ''

		families = del_mp.keys()
		while families:
			family = families.pop()
			afi,safi = family
			mps = del_mp[family]
			addpath = negotiated.addpath.send(*family)
			seen_size = len(packed_del + packed_mp_del)
			mp_packed_generator = MPURNLRI(afi,safi,mps).packed_attributes(addpath,msg_size-seen_size)
			try:
				while True:
					packed = mp_packed_generator.next()
					seen_size = len(packed_del + packed_mp_del + packed)
					if seen_size > msg_size:
						if not packed_mp_del and not packed_del:
							raise Notify(6,0,'attributes size is so large we can not even pack one MPURNLRI')
						yield self._message(Update.prefix(packed_del) + Update.prefix(packed_mp_del))
						packed_del = ''
						packed_mp_del = packed
					else:
						packed_mp_del += packed
			except StopIteration:
				pass

		# add MP

		# we have some MPRNLRI so we need to add the attributes, recalculate
		# and make sure we do not overflow

		packed_mp_add = ''

		if add_mp:
			msg_size = negotiated.msg_size - 19 - 2 - 2 - len(attr)  # 2 bytes for each of the two prefix() header
		seen_size = len(packed_del + packed_mp_del)
		if seen_size > msg_size:
			yield self._message(Update.prefix(packed_del) + Update.prefix(packed_mp_del))
			packed_del = ''
			packed_mp_del = ''

		families = add_mp.keys()
		while families:
			family = families.pop()
			afi,safi = family
			mps = add_mp[family]
			addpath = negotiated.addpath.send(*family)
			seen_size = len(packed_del + packed_mp_del + packed_mp_add)
			mp_packed_generator = MPRNLRI(afi,safi,mps).packed_attributes(addpath,msg_size-seen_size)
			try:
				while True:
					packed = mp_packed_generator.next()
					seen_size = len(packed_del + packed_mp_del + packed_mp_add + packed)
					if seen_size > msg_size:
						if not packed_mp_add and not packed_mp_del and not packed_del:
							raise Notify(6,0,'attributes size is so large we can not even pack on MPURNLRI')
						yield self._message(Update.prefix(packed_del) + Update.prefix(attr + packed_mp_del + packed_mp_add))
						packed_del = ''
						packed_mp_del = ''
						packed_mp_add = packed
					else:
						packed_mp_add += packed
			except StopIteration:
				pass

		# ADD Ipv4

		packed_add = ''

		if add_nlri:
			msg_size = negotiated.msg_size - 19 - 2 - 2 - len(attr)  # 2 bytes for each of the two prefix() header
		seen_size = len(packed_del + packed_mp_del + packed_mp_add)
		if seen_size > msg_size:
			yield self._message(Update.prefix(packed_del) + Update.prefix(packed_mp_del))
			packed_del = ''
			packed_mp_del = ''

		addpath = negotiated.addpath.send(AFI.ipv4,SAFI.unicast)

		while add_nlri:
			nlri = add_nlri.pop()
			packed = nlri.pack(addpath)
			seen_size = len(packed_del + packed_mp_del + packed_mp_add + packed_add + packed)
			if seen_size > msg_size:
				if not packed_add and not packed_mp_add and not packed_mp_del and not packed_del:
					raise Notify(6,0,'attributes size is so large we can not even pack one NLRI')
				if packed_mp_add:
					yield self._message(Update.prefix(packed_del) + Update.prefix(attr + packed_mp_del + packed_mp_add) + packed_add)
					msg_size = negotiated.msg_size - 19 - 2 - 2  # 2 bytes for each of the two prefix() header
				else:
					yield self._message(Update.prefix(packed_del) + Update.prefix(attr + packed_mp_del) + packed_add)
				packed_del = ''
				packed_mp_del = ''
				packed_mp_add = ''
				packed_add = packed
			else:
				packed_add += packed

		yield self._message(Update.prefix(packed_del) + Update.prefix(attr + packed_mp_del + packed_mp_add) + packed_add)

	# XXX: FIXME: this can raise ValueError. IndexError,TypeError, struct.error (unpack) = check it is well intercepted
	@classmethod
	def unpack_message (cls, data, negotiated):
		logger = Logger()

		length = len(data)

		# This could be speed up massively by changing the order of the IF
		if length == 4 and data == '\x00\x00\x00\x00':
			return EOR(AFI.ipv4,SAFI.unicast,IN.ANNOUNCED)  # pylint: disable=E1101
		if length == 11 and data.startswith(EOR.NLRI.PREFIX):
			return EOR.unpack_message(data,negotiated)

		withdrawn, _attributes, announced = cls.split(data)
		attributes = Attributes.unpack(_attributes,negotiated)

		if not withdrawn:
			logger.parser("no withdrawn NLRI")
		if not announced:
			logger.parser("no announced NLRI")

		# Is the peer going to send us some Path Information with the route (AddPath)
		addpath = negotiated.addpath.receive(AFI(AFI.ipv4),SAFI(SAFI.unicast))

		# empty string for NoIP, the packed IP otherwise (without the 3/4 bytes of attributes headers)
		_nexthop = attributes.get(Attribute.CODE.NEXT_HOP,NoIP)
		nexthop = _nexthop.packed

		# XXX: NEXTHOP MUST NOT be the IP address of the receiving speaker.

		nlris = []
		while withdrawn:
			length,nlri = NLRI.unpack(AFI.ipv4,SAFI.unicast,withdrawn,addpath,nexthop,IN.WITHDRAWN)
			logger.parser(LazyFormat("parsed withdraw nlri %s payload " % nlri,withdrawn[:len(nlri)]))
			withdrawn = withdrawn[length:]
			nlris.append(nlri)

		while announced:
			length,nlri = NLRI.unpack(AFI.ipv4,SAFI.unicast,announced,addpath,nexthop,IN.ANNOUNCED)
			logger.parser(LazyFormat("parsed announce nlri %s payload " % nlri,announced[:len(nlri)]))
			announced = announced[length:]
			nlris.append(nlri)

		# required for 'is' comparaison
		UNREACH = [EMPTY_MPURNLRI,]
		REACH = [EMPTY_MPRNLRI,]

		unreach = attributes.pop(MPURNLRI.ID,UNREACH)
		reach = attributes.pop(MPRNLRI.ID,REACH)

		for mpr in unreach:
			nlris.extend(mpr.nlris)

		for mpr in reach:
			nlris.extend(mpr.nlris)

		if not attributes and not nlris:
			# Careful do not use == or != as the comparaison does not work
			if unreach is UNREACH and reach is REACH:
				return EOR(AFI(AFI.ipv4),SAFI(SAFI.unicast))
			if unreach is not UNREACH:
				return EOR(unreach[0].afi,unreach[0].safi)
			if reach is not REACH:
				return EOR(reach[0].afi,reach[0].safi)
			raise RuntimeError('This was not expected')

		return Update(nlris,attributes)



# encoding: utf-8
"""
network.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import random
import socket
import select
from struct import unpack

from exabgp.configuration.environment import environment

from exabgp.util.errstr import errstr

from exabgp.logger import Logger
from exabgp.logger import FakeLogger
from exabgp.logger import LazyFormat

from exabgp.bgp.message import Message

from exabgp.reactor.network.error import error
from exabgp.reactor.network.error import errno
from exabgp.reactor.network.error import NetworkError
from exabgp.reactor.network.error import TooSlowError
from exabgp.reactor.network.error import NotConnected
from exabgp.reactor.network.error import LostConnection
from exabgp.reactor.network.error import NotifyError

from .error import *


class Connection (object):
	direction = 'undefined'
	identifier = 0

	def __init__ (self, afi, peer, local):
		# peer and local are strings of the IP

		try:
			self.defensive = environment.settings().debug.defensive
			self.logger = Logger()
		except RuntimeError:
			self.defensive = True
			self.logger = FakeLogger()

		self.afi = afi
		self.peer = peer
		self.local = local

		self._buffer = ''
		self.io = None
		self.established = False

		self.identifier += 1
		self.id = self.identifier

	# Just in case ..
	def __del__ (self):
		if self.io:
			self.logger.network("%s connection to %s closed" % (self.name(),self.peer),'info')
			self.close()

	def name (self):
		return "session %d %s" % (self.id,self.direction)

	def close (self):
		try:
			self.logger.wire("%s, closing connection from %s to %s" % (self.name(),self.local,self.peer))
			if self.io:
				self.io.close()
				self.io = None
		except KeyboardInterrupt,exc:
			raise exc
		except:
			pass

	def reading (self):
		while True:
			try:
				r,_,_ = select.select([self.io,],[],[],0)
			except select.error,exc:
				if exc.args[0] not in error.block:
					self.close()
					self.logger.wire("%s %s errno %s on socket" % (self.name(),self.peer,errno.errorcode[exc.args[0]]))
					raise NetworkError('errno %s on socket' % errno.errorcode[exc.args[0]])
				return False
			return r != []

	def writing (self):
		while True:
			try:
				_,w,_ = select.select([],[self.io,],[],0)
			except select.error,exc:
				if exc.args[0] not in error.block:
					self.close()
					self.logger.wire("%s %s errno %s on socket" % (self.name(),self.peer,errno.errorcode[exc.args[0]]))
					raise NetworkError('errno %s on socket' % errno.errorcode[exc.args[0]])
				return False
			return w != []

	def _reader (self, number):
		# The function must not be called if it does not return with no data with a smaller size as parameter
		if not self.io:
			self.close()
			raise NotConnected('Trying to read on a closed TCP conncetion')
		if number == 0:
			yield ''
			return

		while not self.reading():
			yield ''
		data = ''
		reported = ''
		while True:
			try:
				while True:
					if self.defensive and random.randint(0,2):
						raise socket.error(errno.EAGAIN,'raising network error in purpose')

					read = self.io.recv(number)
					if not read:
						self.close()
						self.logger.wire("%s %s lost TCP session with peer" % (self.name(),self.peer))
						raise LostConnection('the TCP connection was closed by the remote end')
					data += read

					number -= len(read)
					if not number:
						self.logger.wire(
							LazyFormat(
								"%s %-32s RECEIVED " % (
									self.name(),
									'%s / %s' % (self.local,self.peer)
								),
								read
							)
						)
						yield data
						return

					yield ''
			except socket.timeout,exc:
				self.close()
				self.logger.wire("%s %s peer is too slow" % (self.name(),self.peer))
				raise TooSlowError('Timeout while reading data from the network (%s)' % errstr(exc))
			except socket.error,exc:
				if exc.args[0] in error.block:
					message = "%s %s blocking io problem mid-way through reading a message %s, trying to complete" % (self.name(),self.peer,errstr(exc))
					if message != reported:
						reported = message
						self.logger.wire(message,'debug')
					yield ''
				elif exc.args[0] in error.fatal:
					self.close()
					raise LostConnection('issue reading on the socket: %s' % errstr(exc))
				# what error could it be !
				else:
					self.logger.wire("%s %s undefined error reading on socket" % (self.name(),self.peer))
					raise NetworkError('Problem while reading data from the network (%s)' % errstr(exc))

	def writer (self, data):
		if not self.io:
			# XXX: FIXME: Make sure it does not hold the cleanup during the closing of the peering session
			yield True
			return
		while not self.writing():
			yield False
		self.logger.wire(LazyFormat("%s %-32s SENDING " % (self.name(),'%s / %s' % (self.local,self.peer)),data))
		# The first while is here to setup the try/catch block once as it is very expensive
		while True:
			try:
				while True:
					if self.defensive and random.randint(0,2):
						raise socket.error(errno.EAGAIN,'raising network error in purpose')

					# we can not use sendall as in case of network buffer filling
					# it does raise and does not let you know how much was sent
					number = self.io.send(data)
					if not number:
						self.close()
						self.logger.wire("%s %s lost TCP connection with peer" % (self.name(),self.peer))
						raise LostConnection('lost the TCP connection')

					data = data[number:]
					if not data:
						yield True
						return
					yield False
			except socket.error,exc:
				if exc.args[0] in error.block:
					self.logger.wire(
						"%s %s blocking io problem mid-way through writing a message %s, trying to complete" % (
							self.name(),
							self.peer,
							errstr(exc)
						),
						'debug'
					)
					yield False
				elif exc.errno == errno.EPIPE:
					# The TCP connection is gone.
					self.close()
					raise NetworkError('Broken TCP connection')
				elif exc.args[0] in error.fatal:
					self.close()
					self.logger.wire("%s %s problem sending message (%s)" % (self.name(),self.peer,errstr(exc)))
					raise NetworkError('Problem while writing data to the network (%s)' % errstr(exc))
				# what error could it be !
				else:
					self.logger.wire("%s %s undefined error writing on socket" % (self.name(),self.peer))
					yield False

	def reader (self):
		# _reader returns the whole number requested or nothing and then stops
		for header in self._reader(Message.HEADER_LEN):
			if not header:
				yield 0,0,'','',None

		if not header.startswith(Message.MARKER):
			report = 'The packet received does not contain a BGP marker'
			yield 0,0,header,'',NotifyError(1,1,report)
			return

		msg = ord(header[18])
		length = unpack('!H',header[16:18])[0]

		if length < Message.HEADER_LEN or length > Message.MAX_LEN:
			report = '%s has an invalid message length of %d' % (Message.CODE.name(msg),length)
			yield length,0,header,'',NotifyError(1,2,report)
			return

		validator = Message.Length.get(msg,lambda _: _ >= 19)
		if not validator(length):
			# MUST send the faulty length back
			report = '%s has an invalid message length of %d' % (Message.CODE.name(msg),length)
			yield length,0,header,'',NotifyError(1,2,report)
			return

		number = length - Message.HEADER_LEN

		if not number:
			yield length,msg,header,'',None
			return

		for body in self._reader(number):
			if not body:
				yield 0,0,'','',None

		yield length,msg,header,body,None



# encoding: utf-8
"""
etag.py

Created by Thomas Mangin on 2014-06-26.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

# TODO: take into account E-VPN specs that specify the role of the first bit of ESI
# (since draft-ietf-l2vpn-evpn-05)

from struct import pack
from struct import unpack


class EthernetTag (object):
	MAX = pow(2,32)-1

	__slots__ = ['tag']

	def __init__ (self, tag=0):
		self.tag = tag

	def __str__ (self):
		return repr(self.tag)

	def __repr__ (self):
		return repr(self.tag)

	def pack (self):
		return pack("!L",self.tag)

	def __len__ (self):
		return 4

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.tag != other.tag:
			return -1
		return 0

	def __hash__ (self):
		return hash(self.tag)

	@classmethod
	def unpack (cls, data):
		return cls(unpack("!L",data[:4])[0])



# encoding: utf-8
"""
addpath.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.open.capability.capability import Capability

# ====================================================================== AddPath
#


class AddPath (Capability,dict):
	ID = Capability.CODE.ADD_PATH

	string = {
		0: 'disabled',
		1: 'receive',
		2: 'send',
		3: 'send/receive',
	}

	def __init__ (self, families=(),send_receive=0):
		for afi,safi in families:
			self.add_path(afi,safi,send_receive)

	def add_path (self, afi, safi, send_receive):
		self[(afi,safi)] = send_receive

	def __str__ (self):
		return 'AddPath(' + ','.join(["%s %s %s" % (self.string[self[aafi]],xafi,xsafi) for (aafi,xafi,xsafi) in [((afi,safi),str(afi),str(safi)) for (afi,safi) in self]]) + ')'

	def json (self):
		families = ','.join('"%s/%s": "%s"' % (xafi,xsafi,self.string[self[aafi]]) for (aafi,xafi,xsafi) in (((afi,safi),str(afi),str(safi)) for (afi,safi) in self))
		return '{ "name": "addpath"%s%s }' % (', ' if families else '', families)

	def extract (self):
		rs = ''
		for v in self:
			if self[v]:
				rs += v[0].pack() + v[1].pack() + pack('!B',self[v])
		return [rs,]

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: should check that we have not yet seen the capability
		while data:
			afi = AFI.unpack(data[:2])
			safi = SAFI.unpack(data[2])
			sr = ord(data[3])
			instance.add_path(afi,safi,sr)
			data = data[4:]
		return instance



# encoding: utf-8
"""
section.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised

# ====================================================================== Section
# The common function all Section should have


class Section (object):
	configuration = dict()
	factory = dict()

	# to be defined in subclasses
	name = ''
	syntax = ''

	def __init__ (self):
		# we get our name through our subclasses
		self.factory[self.name] = self

	def drop_parenthesis (self, tokeniser):
		if tokeniser() != '{':
			# syntax is set in our subclasses
			raise Raised(tokeniser,'missing opening parenthesis "{"',self.syntax)

	def create_section (self, section, tokeniser):
		name = tokeniser()
		if name == '{':
			# syntax is set in our subclasses
			raise Raised(tokeniser,'was expecting section name',self.syntax)
		self.drop_parenthesis(tokeniser)
		return self.create_content(section,name,tokeniser)

	def create_content (self, section, name, tokeniser):
		storage = self.configuration[tokeniser.name][section][name]
		if storage:
			# syntax is set in our subclasses
			raise Raised(tokeniser,'the section name %s/%s for %s is not unique' % (section,name,tokeniser.name),self.syntax)
		return storage

	def get_section (self, section, tokeniser):
		name = tokeniser()

		if name == '{':
			tokeniser.rewind(name)
			tokeniser.rewind('anonymous')
			return None

		storage = self.configuration[tokeniser.name][section][name]
		if storage is None:
			# syntax is set in our subclasses
			raise Raised(tokeniser,'the section name %s referenced does not exists' % name,self.syntax)
		return storage

	def extract_anonymous (self, section, tokeniser):
		if 'anonymous' in self.configuration[tokeniser.name][section]:
			storage = self.configuration[tokeniser.name][section]['anonymous']
			del self.configuration[tokeniser.name][section]['anonymous']
			return storage
		else:
			return None

	def _check_duplicate (self, tokeniser, klass):
		# location is set by our caller
		key = self.location[-3]  # pylint: disable=E1101
		if key in self.content:
			raise klass(tokeniser,"duplicate entries for %s" % key)

	# default function for entering and exiting

	def enter (self, tokeniser):
		self.content = self.create_section(self.name,tokeniser)

	def exit (self, tokeniser):
		# no verification to do
		pass

	def enter_nameless (self, tokeniser):
		token = tokeniser()
		if token != '{':
			# syntax is set in our subclasses
			raise Raised(tokeniser,'was expecting {',self.syntax)

	def exit_nameless (self, tokeniser):
		# no verification to do
		pass

	def enter_anonymous (self, tokeniser):
		token = tokeniser()
		if token != '{':
			# syntax is set in our subclasses
			raise Raised(tokeniser,'was expecting {',self.syntax)
		self.content = self.create_content(self.name,'anonymous',tokeniser)

	def exit_anonymous (self, tokeniser):
		# no verification to do
		pass



# encoding: utf-8
"""
pmsi_tunnel.py

Created by Thomas Morin on 2014-06-10.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IPv4
from exabgp.bgp.message.update.attribute.attribute import Attribute

# http://tools.ietf.org/html/rfc6514#section-5
#
#  +---------------------------------+
#  |  Flags (1 octet)                |
#  +---------------------------------+
#  |  Tunnel Type (1 octets)         |
#  +---------------------------------+
#  |  MPLS Label (3 octets)          |
#  +---------------------------------+
#  |  Tunnel Identifier (variable)   |
#  +---------------------------------+


# ========================================================================= PMSI
# RFC 6514

class PMSI (Attribute):
	ID = Attribute.CODE.PMSI_TUNNEL
	FLAG = Attribute.Flag.OPTIONAL | Attribute.Flag.TRANSITIVE
	CACHING = True
	TUNNEL_TYPE = -1

	# TUNNEL_TYPE MUST NOT BE DEFINED HERE ( it allows to set it up as a self. value)

	_pmsi_known = dict()
	_name = {
		0: 'No tunnel',
		1: 'RSVP-TE P2MP LSP',
		2: 'mLDP P2MP LSP',
		3: 'PIM-SSM Tree',
		4: 'PIM-SM Tree',
		5: 'BIDIR-PIM Tree',
		6: 'Ingress Replication',
		7: 'mLDP MP2MP LSP',
	}

	__slots__ = ['label','flags','tunnel']

	def __init__ (self, tunnel, label, flags):
		self.label = label    # integer
		self.flags = flags    # integer
		self.tunnel = tunnel  # tunnel id, packed data

	@staticmethod
	def name (tunnel_type):
		return PMSI._name.get(tunnel_type,'unknown')

	def pack (self):
		return self._attribute(
			pack(
				'!BB3s',
				self.flags,
				self.TUNNEL_TYPE,
				pack('!L',self.label << 4)[1:4]
			) + self.tunnel
		)

	# XXX: FIXME: Orange code had 4 (and another reference to it in the code elsewhere)
	def __len__ (self):
		return len(self.tunnel) + 5  # label:1, tunnel type: 1, MPLS label:3

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		# if self.TUNNEL_TYPE != other.TUNNEL_TYPE:
		# 	return -1
		if self.label != other.label:
			return -1
		if self.flags != other.flags:
			return -1
		if self.tunnel != other.tunnel:
			return -1
		return 0

	def __repr__ (self):
		return str(self)

	def prettytunnel (self):
		return "0x" + ''.join('%02X' % ord(_) for _ in self.tunnel) if self.tunnel else ''

	def __str__ (self):
		# TODO: add hex dump of packedValue
		return "pmsi:%s:%s:%s:%s" % (
			self.name(self.TUNNEL_TYPE).replace(' ','').lower(),
			str(self.flags) if self.flags else '-',  # why not use zero (0) ?
			str(self.label) if self.label else '-',  # what noy use zero (0) ?
			self.prettytunnel()
		)

	@staticmethod
	def register_pmsi (klass):
		PMSI._pmsi_known[klass.TUNNEL_TYPE] = klass

	@staticmethod
	def pmsi_unknown (subtype, tunnel, label, flags):
		pmsi = PMSI(tunnel,label,flags)
		pmsi.TUNNEL_TYPE = subtype
		return pmsi

	@classmethod
	def unpack (cls, data, negotiated):
		flags,subtype = unpack('!BB',data[:2])
		label = unpack('!L','\0'+data[2:5])[0] >> 4
		# should we check for bottom of stack before the shift ?
		if subtype in cls._pmsi_known:
			return cls._pmsi_known[subtype].unpack(data[5:],label,flags)
		return cls.pmsi_unknown(subtype,data[5:],label,flags)


# ================================================================= PMSINoTunnel
# RFC 6514

class PMSINoTunnel (PMSI):
	TUNNEL_TYPE = 0

	def __init__ (self, label=0,flags=0):
		PMSI.__init__(self,'',label,flags)

	def prettytunnel (self):
		return ''

	@classmethod
	def unpack (cls, tunnel, label, flags):
		return cls(label,flags)


# ======================================================= PMSIIngressReplication
# RFC 6514

class PMSIIngressReplication (PMSI):
	TUNNEL_TYPE = 6

	def __init__ (self, ip, label=0,flags=0,tunnel=None):
		self.ip = ip
		PMSI.__init__(self,tunnel if tunnel else IPv4.pton(ip),label,flags)

	def prettytunnel (self):
		return self.ip

	@classmethod
	def unpack (cls, tunnel, label, flags):
		ip = IPv4.ntop(tunnel)
		return cls(ip,label,flags,tunnel)


PMSI.register_pmsi(PMSINoTunnel)
PMSI.register_pmsi(PMSIIngressReplication)



# encoding: utf-8
"""
origin.py

Created by Thomas Mangin on 2014-06-20.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IPv4
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity


# ======================================================================= Origin
# RFC 4360 / RFC 7153

class Origin (ExtendedCommunity):
	COMMUNITY_SUBTYPE = 0x03
	LIMIT = 0  # This is to prevent warnings from scrutinizer

	@property
	def la (self):
		return self.community[2:self.LIMIT]

	@property
	def ga (self):
		return self.community[self.LIMIT:8]


# ================================================================== OriginASNIP
# RFC 4360 / RFC 7153

class OriginASNIP (Origin):
	COMMUNITY_TYPE = 0x00
	LIMIT = 4

	__slots__ = ['asn','ip']

	def __init__ (self, asn, ip, transitive, community=None):
		self.asn = asn
		self.ip = ip
		Origin.__init__(
			self,
			community if community else pack(
				'!2sH4s',
				self._packedTypeSubtype(),
				asn,
				IPv4.pton(ip)
			)
		)

	def __str__ (self):
		return "origin:%s:%s" % (self.asn,self.ip)

	@staticmethod
	def unpack (data):
		asn,ip = unpack('!H4s',data[2:8])
		return OriginASNIP(ASN(asn),IPv4.ntop(ip),False,data[:8])


# ================================================================== OriginIPASN
# RFC 4360 / RFC 7153

class OriginIPASN (Origin):
	COMMUNITY_TYPE = 0x01
	LIMIT = 6

	__slots__ = ['asn','ip']

	def __init__ (self, asn, ip, transitive, community=None):
		self.ip = ip
		self.asn = asn
		Origin.__init__(
			self,
			community if community else pack(
				'!2s4sH',
				self._packedTypeSubtype(),
				IPv4.pton(ip),
				asn
			)
		)

	def __str__ (self):
		return "origin:%s:%s" % (self.ip, self.asn)

	@staticmethod
	def unpack (data):
		ip,asn = unpack('!4sH',data[2:8])
		return OriginIPASN(IPv4.ntop(ip),ASN(asn),False,data[:8])


# ============================================================= OriginASN4Number
# RFC 4360 / RFC 7153

class OriginASN4Number (Origin):
	COMMUNITY_TYPE = 0x02
	LIMIT = 6

	def __init__ (self, asn, number, transitive, community=None):
		self.asn = asn
		self.number = number
		Origin.__init__(
			self,
			community if community else pack(
				'!2sLH',
				self._packedTypeSubtype(),
				asn,
				number
			)
		)

	def __str__ (self):
		return "origin:%sL:%s" % (self.asn, self.number)

	@staticmethod
	def unpack (data):
		asn,number = unpack('!LH',data[2:8])
		return OriginASN4Number(ASN(asn),number,False,data[:8])



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.update.attribute.attribute import Attribute


# =================================================================== Origin (1)

class Origin (Attribute):
	ID = Attribute.CODE.ORIGIN
	FLAG = Attribute.Flag.TRANSITIVE
	CACHING = True

	IGP        = 0x00
	EGP        = 0x01
	INCOMPLETE = 0x02

	__slots__ = ['origin','packed']

	def __init__ (self, origin, packed=None):
		self.origin = origin
		self.packed = self._attribute(packed if packed else chr(origin))

	def pack (self, negotiated=None):
		return self.packed

	def __len__ (self):
		return len(self.packed)

	def __str__ (self):
		if self.origin == 0x00:
			return 'igp'
		if self.origin == 0x01:
			return 'egp'
		if self.origin == 0x02:
			return 'incomplete'
		return 'invalid'

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.origin != other.origin:
			return -1
		return 0

	@classmethod
	def unpack (cls, data, negotiated):
		return cls(ord(data),data)

	@classmethod
	def setCache (cls):
		# there can only be three, build them now
		IGP = Origin(Origin.IGP)
		EGP = Origin(Origin.EGP)
		INC = Origin(Origin.INCOMPLETE)

		cls.cache[Attribute.CODE.ORIGIN][IGP.pack()] = IGP
		cls.cache[Attribute.CODE.ORIGIN][EGP.pack()] = EGP
		cls.cache[Attribute.CODE.ORIGIN][INC.pack()] = INC

Origin.setCache()



# encoding: utf-8
"""
negotiated.py

Created by Thomas Mangin on 2012-07-19.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.asn import AS_TRANS
from exabgp.bgp.message.open.holdtime import HoldTime
from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability import REFRESH
from exabgp.bgp.message.open.routerid import RouterID


class Negotiated (object):
	MAX_SIZE = 4096
	FREE_SIZE = MAX_SIZE - 19 - 2 - 2

	def __init__ (self, neighbor):
		self.neighbor = neighbor

		self.sent_open = None
		self.received_open = None

		self.holdtime = HoldTime(0)
		self.local_as = ASN(0)
		self.peer_as = ASN(0)
		self.families = []
		self.asn4 = False
		self.addpath = RequirePath()
		self.multisession = False
		self.msg_size = self.MAX_SIZE
		self.operational = False
		self.refresh = REFRESH.ABSENT  # pylint: disable=E1101
		self.aigp = None

	def sent (self, sent_open):
		self.sent_open = sent_open
		if self.received_open:
			self._negotiate()

	def received (self, received_open):
		self.received_open = received_open
		if self.sent_open:
			self._negotiate()

	def _negotiate (self):
		sent_capa = self.sent_open.capabilities
		recv_capa = self.received_open.capabilities

		self.holdtime = HoldTime(min(self.sent_open.hold_time,self.received_open.hold_time))

		self.addpath.setup(self.sent_open,self.received_open)
		self.asn4 = sent_capa.announced(Capability.CODE.FOUR_BYTES_ASN) and recv_capa.announced(Capability.CODE.FOUR_BYTES_ASN)
		self.operational = sent_capa.announced(Capability.CODE.OPERATIONAL) and recv_capa.announced(Capability.CODE.OPERATIONAL)

		self.local_as = self.sent_open.asn
		self.peer_as = self.received_open.asn
		if self.received_open.asn == AS_TRANS and self.asn4:
			self.peer_as = recv_capa.get(Capability.CODE.FOUR_BYTES_ASN,self.peer_as)

		self.families = []
		if \
			recv_capa.announced(Capability.CODE.MULTIPROTOCOL) and \
			sent_capa.announced(Capability.CODE.MULTIPROTOCOL):
			for family in recv_capa[Capability.CODE.MULTIPROTOCOL]:
				if family in sent_capa[Capability.CODE.MULTIPROTOCOL]:
					self.families.append(family)

		if recv_capa.announced(Capability.CODE.ENHANCED_ROUTE_REFRESH) and sent_capa.announced(Capability.CODE.ENHANCED_ROUTE_REFRESH):
			self.refresh = REFRESH.ENHANCED  # pylint: disable=E1101
		elif recv_capa.announced(Capability.CODE.ROUTE_REFRESH) and sent_capa.announced(Capability.CODE.ROUTE_REFRESH):
			self.refresh = REFRESH.NORMAL  # pylint: disable=E1101

		self.multisession  = sent_capa.announced(Capability.CODE.MULTISESSION) and recv_capa.announced(Capability.CODE.MULTISESSION)
		self.multisession |= sent_capa.announced(Capability.CODE.MULTISESSION_CISCO) and recv_capa.announced(Capability.CODE.MULTISESSION_CISCO)

		if self.multisession:
			sent_ms_capa = set(sent_capa[Capability.CODE.MULTISESSION])
			recv_ms_capa = set(recv_capa[Capability.CODE.MULTISESSION])

			if sent_ms_capa == set([]):
				sent_ms_capa = set([Capability.CODE.MULTIPROTOCOL])
			if recv_ms_capa == set([]):
				recv_ms_capa = set([Capability.CODE.MULTIPROTOCOL])

			if sent_ms_capa != recv_ms_capa:
				self.multisession = (2,8,'multisession, our peer did not reply with the same sessionid')

			# The way we implement MS-BGP, we only send one MP per session
			# therefore we can not collide due to the way we generate the configuration

			for capa in sent_ms_capa:
				# no need to check that the capability exists, we generated it
				# checked it is what we sent and only send MULTIPROTOCOL
				if sent_capa[capa] != recv_capa[capa]:
					self.multisession = (2,8,'when checking session id, capability %s did not match' % str(capa))
					break

		elif sent_capa.announced(Capability.CODE.MULTISESSION):
			self.multisession = (2,9,'multisession is mandatory with this peer')

		# XXX: Does not work as the capa is not yet defined
		# if received_open.capabilities.announced(Capability.CODE.EXTENDED_MESSAGE) \
		# and sent_open.capabilities.announced(Capability.CODE.EXTENDED_MESSAGE):
		# 	if self.peer.bgp.received_open_size:
		# 		self.received_open_size = self.peer.bgp.received_open_size - 19

	def validate (self, neighbor):
		if self.peer_as != neighbor.peer_as:
			return (2,2,'ASN in OPEN (%d) did not match ASN expected (%d)' % (self.received_open.asn,neighbor.peer_as))

		# RFC 6286 : http://tools.ietf.org/html/rfc6286
		# XXX: FIXME: check that router id is not self
		if self.received_open.router_id == RouterID('0.0.0.0'):
			return (2,3,'0.0.0.0 is an invalid router_id')

		if self.received_open.asn == neighbor.local_as:
			# router-id must be unique within an ASN
			if self.received_open.router_id == neighbor.router_id:
				return (2,3,'BGP Indendifier collision, same router-id (%s) on both side of this IBGP session' % self.received_open.router_id)

		if self.received_open.hold_time and self.received_open.hold_time < 3:
			return (2,6,'Hold Time is invalid (%d)' % self.received_open.hold_time)

		if self.multisession not in (True,False):
			# XXX: FIXME: should we not use a string and perform a split like we do elswhere ?
			# XXX: FIXME: or should we use this trick in the other case ?
			return self.multisession

		return None

# =================================================================== RequirePath


class RequirePath (object):
	REFUSE = 0
	ACCEPT = 1
	ANNOUNCE = 2

	def __init__ (self):
		self._send = {}
		self._receive = {}

	def setup (self, received_open, sent_open):
		# A Dict always returning False
		class FalseDict (dict):
			def __getitem__ (self, key):
				return False

		receive = received_open.capabilities.get(Capability.CODE.ADD_PATH,FalseDict())
		send = sent_open.capabilities.get(Capability.CODE.ADD_PATH,FalseDict())

		# python 2.4 compatibility mean no simple union but using sets.Set
		union = []
		union.extend(send.keys())
		union.extend([k for k in receive.keys() if k not in send.keys()])

		for k in union:
			self._send[k] = bool(receive.get(k,self.REFUSE) & self.ANNOUNCE and send.get(k,self.REFUSE) & self.ACCEPT)
			self._receive[k] = bool(receive.get(k,self.REFUSE) & self.ACCEPT and send.get(k,self.REFUSE) & self.ANNOUNCE)

	def send (self, afi, safi):
		return self._send.get((afi,safi),False)

	def receive (self, afi, safi):
		return self._receive.get((afi,safi),False)



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section


# =================================================================== bgp_syntax

syntax_bgp = """\
bgp {
	lots here
}
"""


# ================================================================ RaisedProcess

class RaisedBGP (Raised):
	syntax = syntax_bgp


# =================================================================== SectionBGP
#

class SectionBGP (Section):
	syntax = syntax_bgp
	name = 'bgp'
	content = None

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		registry.register_hook(cls,'enter',location,'enter_nameless')
		registry.register_hook(cls,'exit',location,'exit_nameless')



"""
mac.py

Created by Thomas Morin on 2014-06-23.
Copyright (c) 2014-2015 Orange. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.labels import Labels
from exabgp.bgp.message.update.nlri.qualifier.esi import ESI
from exabgp.bgp.message.update.nlri.qualifier.etag import EthernetTag

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN


# ===================================================================== EVPNNLRI

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |Ethernet Segment Identifier (10 octets)|
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  MAC Address Length (1 octet)         |
# +---------------------------------------+
# |  MAC Address (6 octets)               |  48 bits is 6 bytes
# +---------------------------------------+
# |  IP Address Length (1 octet)          |  zero if IP Address field absent
# +---------------------------------------+
# |  IP Address (4 or 16 octets)          |
# +---------------------------------------+
# |  MPLS Label (3 octets)                |
# +---------------------------------------+

class MAC (EVPN):
	CODE = 2
	NAME = "MAC/IP advertisement"
	SHORT_NAME = "MACAdv"

	def __init__ (self, rd, esi, etag, mac, maclen, label,ip,packed=None):
		EVPN.__init__(self,packed)
		self.rd = rd
		self.esi = esi
		self.etag = etag
		self.maclen = maclen
		self.mac = mac
		self.ip = ip
		self.label = label if label else Labels.NOLABEL
		self.pack()

	def __str__ (self):
		return "%s:%s:%s:%s:%s:%s%s:%s" % (
			self._prefix(),
			self.rd,
			self.esi,
			self.etag,
			self.mac,
			"" if len(self.mac) == 48 else "/%d" % self.maclen,
			self.ip if self.ip else "",
			self.label
		)

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.rd != other.rd:
			return -1
		# if self.esi == other.esi:  # MUST NOT be part of the test
		# 	return -1
		# if self.label == other.label:  # MUST NOT be part of the test
		# 	return -1
		if self.etag != other.etag:
			return -1
		if self.mac != other.mac:
			return -1
		if self.ip != other.ip:
			return -1
		return 0

	# XXX: FIXME: improve for better performance?
	def __hash__ (self):
		# esi and label MUST *NOT* be part of the hash
		return hash("%s:%s:%s:%s" % (self.rd,self.etag,self.mac,self.ip))

	def pack (self):
		if not self.packed:
			ip = self.ip.pack() if self.ip else ''
			value = "%s%s%s%s%s%s%s%s" % (
				self.rd.pack(),
				self.esi.pack(),
				self.etag.pack(),
				chr(len(self.maclen)),  # will most likely always be 10
				self.mac.pack(),
				chr(len(ip.pack()) if ip else '\x00'),
				ip,
				self.label.pack()
			)
			self.packed = value
		return self.packed

	@classmethod
	def unpack (cls, data):
		rd = RouteDistinguisher.unpack(data[:8])
		esi = ESI.unpack(data[8:18])
		etag = EthernetTag.unpack(data[18:22])
		maclength = ord(data[22])

		if maclength % 8 != 0:
			raise Exception('invalid MAC Address length in %s' % cls.NAME)
		end = 23 + maclength/8

		mac = MAC.unpack(data[23:end])

		length = ord(data[end])
		if length % 8 != 0:
			raise Exception('invalid IP Address length in %s' % cls.NAME)
		iplen = length / 8

		ip = IP.unpack(data[end+1:end+1+iplen])
		label = Labels.unpack(data[end+1+iplen:])

		return cls(rd,esi,etag,mac,length,label,ip,data)



# encoding: utf-8
"""
rtc.py

Created by Thomas Morin on 2014-06-10.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute.community.extended.rt import RouteTarget
from exabgp.protocol.ip.address import Address


class RouteTargetConstraint(Address):
	# XXX: FIXME: no support yet for RTC variable length with prefixing

	__slots__ = ['origin','rt']

	def __init__ (self, afi, safi, origin, rt):
		Address.__init__(self,afi,safi)
		self.origin = origin
		self.rt = rt

	def __len__ (self):
		return (4 + len(self.rt))*8 if self.rt else 1

	def __str__ (self):
		return "rtc %s:%s" % (self.origin,self.rt) if self.rt else "rtc wildcard"

	def __repr__ (self):
		return str(self)

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.origin != other.origin:
			return -1
		if self.rt != other.rt:
			return -1
		return 0

	def __hash__ (self):
		return hash(self.pack())

	def pack (self):
		if self.rt:
			return pack("!BL", len(self), self.origin) + self.rt.pack()
		return pack("!B",0)

	@staticmethod
	def unpack (afi, safi, data):
		length = ord(data[0])

		if length == 0:
			return RouteTargetConstraint(afi,safi,ASN(0),None)

		return RouteTargetConstraint(
			afi,safi,
			ASN(unpack('!L', data[1:5])[0]),
			RouteTarget.unpack(data[5:])
		)



# encoding: utf-8
"""
address.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import socket
from struct import calcsize
from collections import namedtuple

from exabgp.netlink.message import InfoMessage


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |     Length    |     Flags     |    Scope      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Address (InfoMessage):
	class Header (object):
		PACK = '4Bi'
		LEN = calcsize(PACK)

	format = namedtuple('Address', 'family prefixlen flags scope index attributes')

	class Command (object):
		RTM_NEWADDR = 0x14
		RTM_DELADDR = 0x15
		RTM_GETADDR = 0x16

	class Type (object):
		class Family (object):
			AF_INET  = socket.AF_INET
			AF_INET6 = socket.AF_INET6

		class Flag (object):
			IFA_F_SECONDARY  = 0x00  # For secondary address (alias interface)
			IFA_F_PERMANENT  = 0x00  # For a permanent address set by the user.  When this is not set, it means the address was dynamically created (e.g., by stateless autoconfiguration).
			IFA_F_DEPRECATED = 0x00  # Defines deprecated (IPV4) address
			IFA_F_TENTATIVE  = 0x00  # Defines tentative (IPV4) address (duplicate address detection is still in progress)

		class Scope (object):
			RT_SCOPE_UNIVERSE = 0x00  # Global route
			RT_SCOPE_SITE     = 0x00  # Interior route in the local autonomous system
			RT_SCOPE_LINK     = 0x00  # Route on this link
			RT_SCOPE_HOST     = 0x00  # Route on the local host
			RT_SCOPE_NOWHERE  = 0x00  # Destination does not exist

		class Attribute (object):
			IFLA_UNSPEC      = 0x00
			IFLA_ADDRESS     = 0x01
			IFLA_BROADCAST   = 0x02
			IFLA_IFNAME      = 0x03
			IFLA_MTU         = 0x04
			IFLA_LINK        = 0x05
			IFLA_QDISC       = 0x06
			IFLA_STATS       = 0x07
			IFLA_COST        = 0x08
			IFLA_PRIORITY    = 0x09
			IFLA_MASTER      = 0x0A
			IFLA_WIRELESS    = 0x0B
			IFLA_PROTINFO    = 0x0C
			IFLA_TXQLEN      = 0x0D
			IFLA_MAP         = 0x0E
			IFLA_WEIGHT      = 0x0F
			IFLA_OPERSTATE   = 0x10
			IFLA_LINKMODE    = 0x11
			IFLA_LINKINFO    = 0x12
			IFLA_NET_NS_PID  = 0x13
			IFLA_IFALIAS     = 0x14
			IFLA_NUM_VF      = 0x15
			IFLA_VFINFO_LIST = 0x16
			IFLA_STATS64     = 0x17
			IFLA_VF_PORTS    = 0x18
			IFLA_PORT_SELF   = 0x19

	@classmethod
	def getAddresses (cls):
		return cls.extract(Address.Command.RTM_GETADDR)



"""
multicast.py

Created by Thomas Morin on 2014-06-23.
Copyright (c) 2014-2015 Orange. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.etag import EthernetTag

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN


# ===================================================================== EVPNNLRI

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  IP Address Length (1 octet)          |
# +---------------------------------------+
# |   Originating Router's IP Addr        |
# |          (4 or 16 octets)             |
# +---------------------------------------+

class Multicast (EVPN):
	CODE = 3
	NAME = "Inclusive Multicast Ethernet Tag"
	SHORT_NAME = "Multicast"

	def __init__ (self, rd, etag, ip):
		self.rd = rd
		self.etag = etag
		self.ip = ip
		EVPN.__init__(self,self.pack())

	def __str__ (self):
		return "%s:%s:%s:%s" % (
			self._prefix(),
			self.rd,
			self.etag,
			self.ip,
		)

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.rd != other.rd:
			return -1
		if self.etag != other.etag:
			return -1
		if self.ip != other.ip:
			return -1
		return 0

	# XXX: FIXME: improve for better performance?
	def __hash__ (self):
		return hash("%s:%s:%s:%s:%s:%s" % (self.afi,self.safi,self.CODE,self.rd,self.etag,self.ip))

	def pack (self):
		ip = self.ip.pack()
		return '%s%s%s%s' % (
			self.rd.pack(),
			self.etag.pack(),
			chr(len(ip)*8),
			ip
		)

	@classmethod
	def unpack (cls, data):
		rd = RouteDistinguisher.unpack(data[:8])
		etag = EthernetTag.unpack(data[8:12])
		iplen = ord(data[12])*8
		ip = IP.unpack(data[12:12+iplen])
		if iplen not in (4*8,16*8):
			raise Exception("IP len is %d, but EVPN route currently support only IPv4" % iplen)
		return cls(rd,etag,ip)



# encoding: utf-8
"""
nlri/__init__.py

Created by Thomas Mangin on 2013-08-07.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.update.nlri.prefix import Prefix
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.mpls import MPLS
from exabgp.bgp.message.update.nlri.vpls import VPLS
from exabgp.bgp.message.update.nlri.flow import Flow


NLRI.register_nlri(MPLS,AFI.ipv4,SAFI.unicast)
NLRI.register_nlri(MPLS,AFI.ipv6,SAFI.unicast)
NLRI.register_nlri(MPLS,AFI.ipv4,SAFI.multicast)
NLRI.register_nlri(MPLS,AFI.ipv6,SAFI.multicast)

NLRI.register_nlri(MPLS,AFI.ipv4,SAFI.nlri_mpls)
NLRI.register_nlri(MPLS,AFI.ipv6,SAFI.nlri_mpls)
NLRI.register_nlri(MPLS,AFI.ipv4,SAFI.mpls_vpn)
NLRI.register_nlri(MPLS,AFI.ipv6,SAFI.mpls_vpn)

NLRI.register_nlri(VPLS,AFI.l2vpn,SAFI.vpls)

NLRI.register_nlri(Flow,AFI.ipv4,SAFI.flow_ip)
NLRI.register_nlri(Flow,AFI.ipv6,SAFI.flow_ip)
NLRI.register_nlri(Flow,AFI.ipv4,SAFI.flow_vpn)
NLRI.register_nlri(Flow,AFI.ipv6,SAFI.flow_vpn)



# encoding: utf-8
"""
aspath.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack
from struct import error

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.asn import AS_TRANS
from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.notification import Notify


# =================================================================== ASPath (2)
# only 2-4% of duplicated data therefore it is not worth to cache

class ASPath (Attribute):
	AS_SET             = 0x01
	AS_SEQUENCE        = 0x02
	AS_CONFED_SEQUENCE = 0x03
	AS_CONFED_SET      = 0x04
	ASN4               = False

	ID = Attribute.CODE.AS_PATH
	FLAG = Attribute.Flag.TRANSITIVE

	__slots__ = ['as_seq','as_set','as_cseq','as_cset','segments','packed','index','_str','_json']

	def __init__ (self, as_sequence, as_set, as_conf_sequence=None,as_conf_set=None,index=None):
		self.as_seq = as_sequence
		self.as_set = as_set
		self.as_cseq = as_conf_sequence if as_conf_sequence is not None else []
		self.as_cset = as_conf_set if as_conf_set is not None else []
		self.segments = ''
		self.packed = {True:'',False:''}
		self.index = index  # the original packed data, use for indexing
		self._str = ''
		self._json = {}

	def __cmp__ (self, other):
		if not isinstance(other, self.__class__):
			return -1
		if self.ASN4 != other.ASN4:
			return -1
		if self.as_seq != other.as_seq:
			return -1
		if self.as_set != other.as_set:
			return -1
		return 0

	def _segment (self, seg_type, values, asn4):
		length = len(values)
		if length:
			if length > 255:
				return self._segment(seg_type,values[:255],asn4) + self._segment(seg_type,values[255:],asn4)
			return "%s%s%s" % (chr(seg_type),chr(len(values)),''.join([v.pack(asn4) for v in values]))
		return ""

	def _segments (self, asn4):
		segments = ''
		if self.as_cseq:
			segments += self._segment(self.AS_CONFED_SEQUENCE,self.as_cseq,asn4)
		if self.as_cset:
			segments += self._segment(self.AS_CONFED_SET,self.as_cset,asn4)
		if self.as_seq:
			segments += self._segment(self.AS_SEQUENCE,self.as_seq,asn4)
		if self.as_set:
			segments += self._segment(self.AS_SET,self.as_set,asn4)
		return segments

	def asn_pack (self, negotiated, force_asn4=False):
		asn4 = True if force_asn4 else negotiated.asn4
		if not self.packed[asn4]:
			self.packed[asn4] = self._attribute(self._segments(asn4))
		return self.packed[asn4]

	def pack (self, negotiated):
		# if the peer does not understand ASN4, we need to build a transitive AS4_PATH
		if negotiated.asn4:
			return self.asn_pack(negotiated)

		as2_seq = [_ if not _.asn4() else AS_TRANS for _ in self.as_seq]
		as2_set = [_ if not _.asn4() else AS_TRANS for _ in self.as_set]

		message = ASPath(as2_seq,as2_set,self.as_cseq,self.as_cset).asn_pack(negotiated,False)
		if AS_TRANS in as2_seq or AS_TRANS in as2_set:
			message += AS4Path(self.as_seq,self.as_set,self.as_cseq,self.as_cset).asn_pack(negotiated,True)
		return message

	def __len__ (self):
		raise RuntimeError('it makes no sense to ask for the size of this object')

	def __str__ (self, confed=False):
		if self._str:
			return self._str

		if self.as_cseq or self.as_cset:
			string = self.string(self.as_seq,self.as_set) + self.string(self.as_cseq,self.as_cset)
		else:
			string = self.string(self.as_seq,self.as_set)

		self._str = string
		return string

	def string (self, aseq, aset):
		lseq = len(aseq)
		lset = len(aset)
		if lseq == 1:
			if not lset:
				string = '[ %d ]' % aseq[0]
			else:
				string = '[ %s %s]' % (aseq[0],'( %s ) ' % (' '.join([str(_) for _ in aset])))
		elif lseq > 1:
			if lset:
				string = '[ %s %s]' % ((' '.join([str(_) for _ in aseq])),'( %s ) ' % (' '.join([str(_) for _ in aset])))
			else:
				string = '[ %s ]' % ' '.join([str(_) for _ in aseq])
		else:  # lseq == 0
			if lset:
				string = '[ ( %s )]' % (' '.join([str(_) for _ in aset]))
			else:
				string = '[ ]'
		return string

	def json (self, name):
		match = {
			# data , default representation
			'as-path':            (self.as_seq,  '[]'),
			'as-set':             (self.as_set,  ''),
			'confederation-path': (self.as_cseq, '[]'),
			'confederation-set':  (self.as_cset, ''),
		}

		data,default = match[name]
		self._json[name] = '[ %s ]' % ', '.join([str(_) for _ in data]) if data else default
		return self._json[name]

	@classmethod
	def _new_aspaths (cls, data, asn4, klass=None):
		as_set = []
		as_seq = []
		as_cset = []
		as_cseq = []

		backup = data

		unpacker = {
			False: '!H',
			True:  '!L',
		}
		size = {
			False: 2,
			True:  4,
		}
		as_choice = {
			ASPath.AS_SEQUENCE: as_seq,
			ASPath.AS_SET:      as_set,
			ASPath.AS_CONFED_SEQUENCE: as_cseq,
			ASPath.AS_CONFED_SET:      as_cset,
		}

		upr = unpacker[asn4]
		length = size[asn4]

		try:

			while data:
				stype = ord(data[0])
				slen  = ord(data[1])

				if stype not in (ASPath.AS_SET, ASPath.AS_SEQUENCE, ASPath.AS_CONFED_SEQUENCE, ASPath.AS_CONFED_SET):
					raise Notify(3,11,'invalid AS Path type sent %d' % stype)

				end = 2+(slen*length)
				sdata = data[2:end]
				data = data[end:]
				# Eat the data and ignore it if the ASPath attribute is know known
				asns = as_choice.get(stype,[])

				for _ in range(slen):
					asn = unpack(upr,sdata[:length])[0]
					asns.append(ASN(asn))
					sdata = sdata[length:]

		except IndexError:
			raise Notify(3,11,'not enough data to decode AS_PATH or AS4_PATH')
		except error:  # struct
			raise Notify(3,11,'not enough data to decode AS_PATH or AS4_PATH')

		if klass:
			return klass(as_seq,as_set,as_cseq,as_cset,backup)
		return cls(as_seq,as_set,as_cseq,as_cset,backup)

	@classmethod
	def unpack (cls, data, negotiated):
		if not data:
			return None  # ASPath.Empty
		return cls._new_aspaths(data,negotiated.asn4,ASPath)


ASPath.Empty = ASPath([],[])


# ================================================================= AS4Path (17)
#

class AS4Path (ASPath):
	ID = Attribute.CODE.AS4_PATH
	FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL
	ASN4 = True

	def pack (self, negotiated=None):
		ASPath.pack(self,True)

	@classmethod
	def unpack (cls, data, negotiated):
		if not data:
			return None  # AS4Path.Empty
		return cls._new_aspaths(data,True,AS4Path)

AS4Path.Empty = AS4Path([],[])



# encoding: utf-8
"""
interface.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import socket
from struct import pack
from struct import unpack
from struct import calcsize
from collections import namedtuple

from exabgp.netlink import NetLinkError
from exabgp.netlink.sequence import Sequence
from exabgp.netlink.attributes import Attributes


try:
	getattr(socket,'AF_NETLINK')
except AttributeError:
	raise ImportError('This module only works on unix version with netlink support')


class NetLinkMessage (object):
	_IGNORE_SEQ_FAULTS = True

	NETLINK_ROUTE = 0

	format = namedtuple('Message','format_type control_flags sequence pid data')
	netlink = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW, NETLINK_ROUTE)

	class Header (object):
		# linux/netlink.h
		PACK  = 'IHHII'
		LEN = calcsize(PACK)

	class Command (object):
		NLMSG_NOOP    = 0x01
		NLMSG_ERROR   = 0x02
		NLMSG_DONE    = 0x03
		NLMSG_OVERRUN = 0x04

	class Flags (object):
		NLM_F_REQUEST = 0x01  # It is query message.
		NLM_F_MULTI   = 0x02  # Multipart message, terminated by NLMSG_DONE
		NLM_F_ACK     = 0x04  # Reply with ack, with zero or error code
		NLM_F_ECHO    = 0x08  # Echo this query

		# Modifiers to GET query
		NLM_F_ROOT   = 0x100  # specify tree root
		NLM_F_MATCH  = 0x200  # return all matching
		NLM_F_DUMP   = NLM_F_ROOT | NLM_F_MATCH
		NLM_F_ATOMIC = 0x400  # atomic GET

		# Modifiers to NEW query
		NLM_F_REPLACE = 0x100  # Override existing
		NLM_F_EXCL    = 0x200  # Do not touch, if it exists
		NLM_F_CREATE  = 0x400  # Create, if it does not exist
		NLM_F_APPEND  = 0x800  # Add to end of list

	errors = {
		Command.NLMSG_ERROR:   'netlink error',
		Command.NLMSG_OVERRUN: 'netlink overrun',
	}

	@classmethod
	def encode (cls, format_type, sequence, control_flags, pid, body, attributes):
		attrs = Attributes.encode(attributes)
		length = cls.Header.LEN + len(attrs) + len(body)
		return pack(cls.Header.PACK, length, format_type, control_flags, sequence, pid) + body + attrs

	@classmethod
	def decode (cls, data):
		while data:
			length, format_type, control_flags, sequence, pid = unpack(cls.Header.PACK,data[:cls.Header.LEN])
			if len(data) < length:
				raise NetLinkError("Buffer underrun")
			yield cls.format(format_type, control_flags, sequence, pid, data[cls.Header.LEN:length])
			data = data[length:]

	# pack('Bxxx', family),
	# family=socket.AF_UNSPEC,

	@classmethod
	def send (cls, format_type, control_flags, family=socket.AF_UNSPEC, attributes=None):
		sequence = Sequence()
		pid = os.getpid()

		if attributes is None:
			attributes = {}

		message = cls.encode(
			format_type,
			sequence,
			control_flags,
			pid,
			pack('Bxxx', family),
			attributes
		)

		cls.netlink.send(message)

		while True:
			response = cls.netlink.recv(640000)
			for response_type, response_flags, response_sequence, response_pid, response_data in cls.decode(response):
				if response_sequence != sequence:
					if cls._IGNORE_SEQ_FAULTS:
						continue
					raise NetLinkError("netlink sequence mismatch")
				elif response_pid != pid:
					raise NetLinkError("netlink pid mismatch")
				# elif response_flags != control_flags:
				# 	raise NetLinkError("netlink flags mismatch")
				elif response_type in cls.errors:
					raise NetLinkError(cls.errors[response_type],message,response)

				if response_type == NetLinkMessage.Command.NLMSG_DONE:
					raise StopIteration()
				yield response_data


class InfoMessage (object):
	# to be defined by the subclasses
	format = namedtuple('Parent', 'to be subclassed')

	DEFAULT_FLAGS = NetLinkMessage.Flags.NLM_F_REQUEST | NetLinkMessage.Flags.NLM_F_DUMP

	# to be defined by the subclasses
	class Header (object):
		PACK = ''
		LEN = 0

	@classmethod
	def decode (cls, data):
		extracted = list(unpack(cls.Header.PACK,data[:cls.Header.LEN]))
		attributes = Attributes.decode(data[cls.Header.LEN:])
		extracted.append(dict(attributes))
		return cls.format(*extracted)

	@classmethod
	def extract (cls, format_type, control_flags=DEFAULT_FLAGS, family=socket.AF_UNSPEC, attributes=None):
		for data in NetLinkMessage.send(format_type,control_flags,family,attributes):
			yield cls.decode(data)



# encoding: utf-8
"""
asn.py

Created by Thomas Mangin on 2015-01-15.
Copyright (c) 2015-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section
from exabgp.configuration.engine.parser import asn


# =================================================================== syntax_asn

syntax_asn = """\
asn {
	local <asn>
	peer  <asn>
}
"""


# ==================================================================== RaisedASN

class RaisedASN (Raised):
	syntax = syntax_asn


# =================================================================== ASNSession
#

class SectionASN (Section):
	syntax = syntax_asn
	name = 'asn'

	def enter_asn (self, tokeniser):
		Section.enter(self,tokeniser)

	def exit_asn (self, tokeniser):
		if 'local' not in self.content:
			raise RaisedASN(tokeniser,'section is missing a local asn')

		if 'peer' not in self.content:
			raise RaisedASN(tokeniser,'section is missing a peer asn')

	def local_asn (self, tokeniser):
		try:
			self.content['local'] = asn(tokeniser)
		except ValueError,exc:
			raise RaisedASN(tokeniser,'could not parse local asn, %s' % str(exc))

	def peer_asn (self, tokeniser):
		try:
			self.content['peer'] = asn(tokeniser)
		except ValueError,exc:
			raise RaisedASN(tokeniser,'could not parse peer asn, %s' % str(exc))

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		registry.register_hook(cls,'enter',location,'enter_asn')
		registry.register_hook(cls,'action',location+['local'],'local_asn')
		registry.register_hook(cls,'action',location+['peer'],'peer_asn')
		registry.register_hook(cls,'exit', location,'exit_asn')



# encoding: utf-8
"""
enum.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


class Enum (int):
	VALUE = {}
	NAME = {}
	UNKNOWN = {}

	@classmethod
	def Name (klass,name):
		lower = name.lower()

		if lower in klass.VALUE:
			return klass.VALUE[lower]

		if lower.isdigit():
			code = klass(lower)
			if code in klass.NAME:
				return klass.VALUE[klass.NAME[code]]

		raise ValueError(klass.UNKNOWN % lower)

	def __str__ (self):
		return self.NAME.get(self,self.UNKNOWN % ('%d' % self))



# encoding: utf-8
"""
tokeniser.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.util import coroutine
from exabgp.configuration.engine.location import Location
from exabgp.configuration.engine.raised import Raised

# convert special caracters


@coroutine.join
def unescape (string):
	start = 0
	while start < len(string):
		pos = string.find('\\', start)
		if pos == -1:
			yield string[start:]
			break
		yield string[start:pos]
		pos += 1
		esc = string[pos]
		if esc == 'b':
			yield '\b'
		elif esc == 'f':
			yield '\f'
		elif esc == 'n':
			yield '\n'
		elif esc == 'r':
			yield '\r'
		elif esc == 't':
			yield '\t'
		elif esc == 'u':
			yield chr(int(string[pos + 1:pos + 5], 16))
			pos += 4
		else:
			yield esc
		start = pos + 1


# A coroutine which return the producer token, or string if quoted from the stream

@coroutine.each
def tokens (stream):
	spaces = [' ','\t','\r','\n']
	strings = ['"', "'"]
	syntax = [',','[',']','{','}']
	comment = ['#',]
	nb_lines = 0
	for line in stream:
		nb_lines += 1
		nb_chars = 0
		quoted = ''
		word = ''
		for char in line:
			if char in comment:
				if quoted:
					word += char
					nb_chars += 1
				else:
					if word:
						yield nb_lines,nb_chars,line,char
						word = ''
					break

			elif char in syntax:
				if quoted:
					word += char
				else:
					if word:
						yield nb_lines,nb_chars-len(word),line,word
						word = ''
					yield nb_lines,nb_chars,line,char
				nb_chars += 1

			elif char in spaces:
				if quoted:
					word += char
				elif word:
					yield nb_lines,nb_chars-len(word),line,word
					word = ''
				nb_chars += 1

			elif char in strings:
				word += char
				if quoted == char:
					quoted = ''
					yield nb_lines,nb_chars-len(word),line,word
					word = ''
				else:
					quoted = char
				nb_chars += 1

			else:
				word += char
				nb_chars += 1

# ==================================================================== Tokeniser
# Return the producer token from the configuration


class Tokeniser (Location):
	def __init__ (self, name, stream):
		super(Tokeniser,self).__init__()
		self.name = name                  # A unique name for this tokenier, so we can have multiple
		self.tokeniser = tokens(stream)   # A corouting giving us the producer toker
		self._rewind = []                 # Should we want to rewind, the list of to pop first

	def __call__ (self):
		if self._rewind:
			return self._rewind.pop()
		token = self.content(self.tokeniser)
		return token

	# XXX: FIXME: line and position only work if we only rewind one element
	def rewind (self, token):
		self._rewind.append(token)

	def content (self, producer):
		try:
			while True:
				self.idx_line,self.idx_column,self.line,token = producer()
				if token == '[':
					returned = []
					for token in self.iterate_list(producer):
						returned.append((self.idx_line,self.idx_column,self.line,token))
					return returned
				elif token[0] in ('"',"'"):
					return unescape(token[1:-1])
				else:
					return token
		except ValueError:
			raise Raised(Location(self.idx_line,self.idx_column,self.line),'Could not parse %s' % str(token))
		except StopIteration:
			return None

	def iterate_list (self, producer):
		token = self.content(producer)
		while token and token != ']':
			yield token
			token = self.content(producer)



# encoding: utf-8
"""
ordereddict.py

Created by Thomas Mangin on 2013-03-18.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


# ================================================================== OrderedDict
# This is only an hack until we drop support for python version < 2.7

class OrderedDict (dict):
	def __init__(self, args):
		dict.__init__(self, args)
		self._order = [_ for _,__ in args]

	def __setitem__(self, key, value):
		dict.__setitem__(self, key, value)
		if key in self._order:
			self._order.remove(key)
		self._order.append(key)

	def __delitem__(self, key):
		dict.__delitem__(self, key)
		self._order.remove(key)

	def keys (self):
		return self._order

	def __iter__ (self):
		return self.__next__()

	def __next__ (self):
		for order in self._order:
			yield order

if __name__ == '__main__':
	d = OrderedDict(((10,'ten'),(8,'eight'),(6,'six'),(4,'four'),(2,'two'),(0,'boom')))
	for k in d:
		print k



# encoding: utf-8
"""
keepalive.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message import Message

# =================================================================== KeepAlive
#


class KeepAlive (Message):
	ID = Message.CODE.KEEPALIVE
	TYPE = chr(Message.CODE.KEEPALIVE)

	def message (self):
		return self._message('')

	def __str__ (self):
		return "KEEPALIVE"

	@classmethod
	def unpack_message (cls, data, negotiated):  # pylint: disable=W0613
		# XXX: FIXME: raise Notify if data has something
		return cls()



# encoding: utf-8
"""
encapsulation.py

Created by Thomas Mangin on 2014-06-20.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity

# ================================================================ Encapsulation
# RFC 5512


class Encapsulation (ExtendedCommunity):
	COMMUNITY_TYPE = 0x03
	COMMUNITY_SUBTYPE = 0x0C

	# https://www.iana.org/assignments/bgp-parameters/bgp-parameters.xhtml#tunnel-types
	class Type:
		DEFAULT   = 0x00
		L2TPv3    = 0x01
		GRE       = 0x02
		IPIP      = 0x07
		VXLAN     = 0x08
		NVGRE     = 0x09
		MPLS      = 0x10
		VXLAN_GPE = 0x0C
		MPLS_UDP  = 0x0D

	_string = {
		Type.DEFAULT:   "Default",
		Type.L2TPv3:    "L2TPv3",
		Type.GRE:       "GRE",
		Type.IPIP:      "IP-in-IP",
		Type.VXLAN:     "VXLAN",
		Type.NVGRE:     "NVGRE",
		Type.MPLS:      "MPLS",
		Type.VXLAN_GPE: "VXLAN-GPE",
		Type.MPLS_UDP:  "MPLS-in-UDP",
	}

	__slots__ = ['tunnel_type']

	def __init__ (self, tunnel_type, community=None):
		self.tunnel_type = tunnel_type
		ExtendedCommunity.__init__(
			self,community if community is not None else pack(
				"!2sLH",
				self._packedTypeSubtype(),
				0,self.tunnel_type
			)
		)

	def __str__ (self):
		return "Encapsulation: %s" % Encapsulation._string.get(self.tunnel_type,"Encap:(unknown:%d)" % self.tunnel_type)

	@staticmethod
	def unpack (data):
		tunnel, = unpack('!H',data[6:8])
		return Encapsulation(tunnel,data[:8])

		# type_  = ord(data[0]) & 0x0F
		# stype = ord(data[1])

		# assert(type_==Encapsulation.COMMUNITY_TYPE)
		# assert(stype==Encapsulation.COMMUNITY_SUBTYPE)
		# assert(len(data)==6)



# encoding: utf-8
"""
debug.py

Created by Thomas Mangin on 2011-03-29.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import sys
from exabgp.util.panic import panic
from exabgp.util.panic import footer


def bug_report (dtype, value, trace):
	print panic

	import traceback

	print "-- Traceback\n\n"
	traceback.print_exception(dtype,value,trace)

	from exabgp.logger import Logger
	logger = Logger()

	print "\n\n-- Configuration\n\n"
	print logger.config()
	print "\n\n-- Logging History\n\n"
	print logger.history()
	print "\n\n\n"

	print footer


def intercept (dtype, value, trace):
	bug_report(dtype, value, trace)
	if os.environ.get('PDB',None) not in [None,'0','']:
		import pdb
		pdb.pm()


def setup_report ():
	sys.excepthook = intercept



# encoding: utf-8
"""
check.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import copy

# common

from exabgp.protocol.family import known_families
from exabgp.bgp.message.update import Update
from exabgp.bgp.message.open import Open
from exabgp.bgp.message.open.capability.capabilities import Capabilities
from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability.negotiated import Negotiated
from exabgp.bgp.message.notification import Notify

# check_neighbor

from exabgp.util.od import od
from exabgp.rib.change import Change

# check_update

from exabgp.reactor.peer import Peer
from exabgp.reactor.api.encoding import JSON

# check_notification

from exabgp.bgp.message.notification import Notification

from exabgp.protocol.family import known_families

# =============================================================== check_neighbor
# ...


def check_neighbor (neighbors):
	from exabgp.logger import Logger

	logger = Logger()
	logger._option.parser = True

	logger.parser('\ndecoding routes in configuration')

	for name in neighbors.keys():
		neighbor = copy.deepcopy(neighbors[name])

		path = {}
		for f in known_families():
			if neighbor.add_path:
				path[f] = neighbor.add_path

		capa = Capabilities().new(neighbor,False)
		if path:
			capa[Capability.CODE.ADD_PATH] = path
		capa[Capability.CODE.MULTIPROTOCOL] = neighbor.families()

		o1 = Open(4,neighbor.local_as,'127.0.0.2',capa,180)
		o2 = Open(4,neighbor.peer_as,'127.0.0.3',capa,180)
		negotiated = Negotiated(neighbor)
		negotiated.sent(o1)
		negotiated.received(o2)
		# grouped = False

		for message in neighbor.rib.outgoing.updates(False):
			pass

		for change1 in neighbor.rib.outgoing.sent_changes():
			str1 = change1.extensive()
			packed = list(Update([change1.nlri],change1.attributes).messages(negotiated))
			pack1 = packed[0]

			logger.parser('parsed route requires %d updates' % len(packed))
			logger.parser('parsed route requires %d updates' % len(packed))
			logger.parser('update size is %d' % len(pack1))

			logger.parser('parsed route %s' % str1)
			logger.parser('parsed hex   %s' % od(pack1))

			# This does not take the BGP header - let's assume we will not break that :)
			try:
				logger.parser('')  # new line

				pack1s = pack1[19:] if pack1.startswith('\xFF'*16) else pack1
				update = Update.unpack_message(pack1s,negotiated)

				change2 = Change(update.nlris[0],update.attributes)
				str2 = change2.extensive()
				pack2 = list(Update([update.nlris[0]],update.attributes).messages(negotiated))[0]

				logger.parser('recoded route %s' % str2)
				logger.parser('recoded hex   %s' % od(pack2))

				str1r = str1.replace(' med 100','').replace(' local-preference 100','').replace(' origin igp','')
				str2r = str2.replace(' med 100','').replace(' local-preference 100','').replace(' origin igp','')

				if ' name ' in str1r:
					parts = str1r.split(' ')
					pos = parts.index('name')
					str1r = ' '.join(parts[:pos] + parts[pos+2:])

				skip = False

				if str1r != str2r:
					if 'attribute [' in str1r and ' 0x00 ' in str1r:
						# we do not decode non-transitive attributes
						logger.parser('skipping string check on update with non-transitive attribute(s)')
						skip = True
					else:
						logger.parser('strings are different:')
						logger.parser('[%s]' % (str1r))
						logger.parser('[%s]' % (str2r))
						return False
				else:
					logger.parser('strings are fine')

				if skip:
					logger.parser('skipping encoding for update with non-transitive attribute(s)')
				elif pack1 != pack2:
					logger.parser('encoding are different')
					logger.parser('[%s]' % (od(pack1)))
					logger.parser('[%s]' % (od(pack2)))
					return False
				else:
					logger.parser('encoding is fine')
					logger.parser('----------------------------------------')

				logger.parser('JSON nlri %s' % change1.nlri.json())
				logger.parser('JSON attr %s' % change1.attributes.json())

			except Notify,exc:
				logger.parser('----------------------------------------')
				logger.parser(str(exc))
				logger.parser('----------------------------------------')
				return False
		neighbor.rib.clear()

	return True


# ================================================================ check_message
#

def check_message (neighbor, message):
	message = message.replace(':','')
	raw = ''.join(chr(int(_,16)) for _ in (message[i*2:(i*2)+2] for i in range(len(message)/2)))

	if raw.startswith('\xff'*16):
		kind = ord(raw[18])
		# XXX: FIXME: check size
		# size = (ord(raw[16]) << 16) + (ord(raw[17]))

		if kind == 1:
			return check_open(neighbor,raw[18:])
		elif kind == 2:
			return check_update(neighbor,raw)
		elif kind == 3:
			return check_notification(raw)
	else:
		return check_update(neighbor,raw)


# ================================================================= check_update
#

def check_open (neighbor, raw):
	pass


# ================================================================= check_update
#

def check_update (neighbor, raw):
	from exabgp.logger import Logger

	logger = Logger()
	logger._parser = True
	logger.parser('\ndecoding routes in configuration')

	n = neighbor[neighbor.keys()[0]]
	p = Peer(n,None)

	path = {}
	for f in known_families():
		if n.add_path:
			path[f] = n.add_path

	capa = Capabilities().new(n,False)
	capa[Capability.CODE.ADD_PATH] = path
	capa[Capability.CODE.MULTIPROTOCOL] = n.families()

	routerid_1 = str(n.router_id)
	routerid_2 = '.'.join(str((int(_)+1) % 250) for _ in str(n.router_id).split('.',-1))

	o1 = Open(4,n.local_as,routerid_1,capa,180)
	o2 = Open(4,n.peer_as,routerid_2,capa,180)
	negotiated = Negotiated(n)
	negotiated.sent(o1)
	negotiated.received(o2)
	# grouped = False

	while raw:
		if raw.startswith('\xff'*16):
			kind = ord(raw[18])
			size = (ord(raw[16]) << 16) + (ord(raw[17]))

			injected,raw = raw[19:size],raw[size:]

			if kind == 2:
				logger.parser('the message is an update')
				decoding = 'update'
			else:
				logger.parser('the message is not an update (%d) - aborting' % kind)
				return False
		else:
			logger.parser('header missing, assuming this message is ONE update')
			decoding = 'update'
			injected,raw = raw,''

		try:
			# This does not take the BGP header - let's assume we will not break that :)
			update = Update.unpack_message(injected,negotiated)
		except KeyboardInterrupt:
			raise
		except Notify,exc:
			logger.parser('could not parse the message','error')
			logger.parser(str(exc),'error')
			return False
		except Exception,exc:
			logger.parser('could not parse the message','error')
			logger.parser(str(exc),'error')
			return False

		logger.parser('')  # new line
		for number in range(len(update.nlris)):
			change = Change(update.nlris[number],update.attributes)
			logger.parser('decoded %s %s %s' % (decoding,change.nlri.action,change.extensive()))
		logger.parser('update json %s' % JSON('3.4.0').update(p,update,'',''))

	return True


# ================================================================= check_update
#

def check_notification (raw):
	notification = Notification.unpack_message(raw[18:],None)
	# XXX: FIXME: should be using logger here
	print notification
	return True



# encoding: utf-8
"""
nexthop.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.protocol.ip import NoIP
from exabgp.bgp.message.update.attribute.attribute import Attribute


# ================================================================== NextHop (3)

# The inheritance order is important and attribute MUST be first for the righ register to be called
# At least until we rename them to be more explicit

class NextHop (Attribute,IP):
	ID = Attribute.CODE.NEXT_HOP
	FLAG = Attribute.Flag.TRANSITIVE
	CACHING = True

	def __init__ (self, ip, packed=None):
		# Need to conform to from IP interface
		self.ip = ip
		self.packed = packed if packed else IP.create(ip).pack()

	def pack (self, negotiated=None):
		return self._attribute(self.packed)

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.pack() != other.pack():
			return -1
		return 0

	@classmethod
	def unpack (cls, data, negotiated=None):
		if not data:
			return NoIP
		return IP.unpack(data,NextHop)

	def __str__ (self):
		return IP.__str__(self)
